#!/bin/env python

"""core.py: Contains the main class of sexiibot"""

__author__ = "raspbeguy"
__copyright__ = "Copyright 2016, raspbeguy"
__credits__ = ["raspbeguy"]
__license__ = "WTFPL"
__version__ = "0.1.1"
__maintainer__ = "raspbeguy"
__email__ = "raspbeguy@hashtagueule.fr"
__status__ = "Devlopment"

from sexiibot.channel import Channel
import os
import subprocess
import time
import threading
import shutil
import socket
import struct

class Sexiibot(object):
    """Core sexiibot class"""

    def __init__(self, server, nick, new=True, port=None, ssl=False, channels=[], realname=None, iipath=None, sock="/var/tmp/%s/sexiibot.sock" % os.environ["USER"]):
        if new:
            self.__server   = server
        else:
            self.__server   = None
        self.__port         = port
        self.__ssl          = ssl
        self.__channels     = {}
        self.__nick         = nick
        if not realname is None:
            self.__realname = realname
        else:
            self.__realname = nick
        self.__in           = None
        self.__out          = None
        self.__iipath       = iipath
        self.__iiproc       = None
        for chan in channels:
            self.__channels[chan] = Channel(chan,self)
        self.__treated      = channels
        self.__quit         = False
        self.__watch        = None
        sockdir = os.path.dirname(sock)
        if not os.path.exists(sockdir):
            os.makedirs(sockdir)
        self.__sock         = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            os.unlink(sock)
        except OSError:
            if os.path.exists(sock):
                raise
        self.__sock.bind(sock)

    def start(self):
        if os.path.exists(self.getIrcPath()):
            for dirname in next(os.walk(self.getIrcPath()))[1]:
                print ("deleting " + dirname)
                shutil.rmtree(self.getIrcPath() + dirname)
        if not self.__server is None:
            try:
                os.remove(self.getIrcPath()+"in")
            except OSError:
                pass
            self.__watch = threading.Thread(target=self.watchChan)
            self.__watch.start()
            if self.__port is None:
                if self.__ssl:
                    self.__port = 6697
                else:
                    self.__port = 6667
            cmd = [self.__iipath, "-s", self.__server, "-p", "%d" % self.__port, "-n", self.__nick, "-f", self.__realname]
            if self.__ssl:
                cmd += ["-e", "ssl"]
            print(" ".join(cmd))
            self.__iiproc = subprocess.Popen(cmd) # self.__iiproc.terminate() later to stop
            time.sleep(2) #temporaire
        self.__in = open(self.getIrcPath() + "in", 'w', 1)
        self.__out = open(self.getIrcPath() + "out", 'r')
        chans_t0 = list(self.__channels.values())
        for chan in chans_t0:
            chan.join()
        self.listenCommands()

    def messageServer(self, message):
        self.__in.write(message + "\n")
        print("Command to server: %s" % message)

    def sendJoin(self, channel):
        self.messageServer("/j %s" % channel)

    def messageChan(self, args):
        if len(args) == 2:
            self.__channels[args[0]].message(args[1])

    def actionChan(self, args):
        if len(args) == 3:
            self.__channels[args[0]].action(args[1])

    def getIrcPath(self):
        return "%s/irc/%s/" % (os.environ["HOME"], self.__server)

    def getNick(self):
        return self.__nick

    def setNick(self, nick):
        self.messageServer("/n %s" % nick)
        self.__nick = nick

    def newChan(self, args):
        if len(args) > 0:
            if args[0] in self.__channels:
                return "warning: already joined %s" % args[0]
            chan = Channel(args[0],self)
            self.__channels[args[0]] = chan
            chan.join()
            response = ""
            if len(args) > 1:
                response += "warning: \"" + ' '.join(args[1:]) + "\" garbage\n"
            response += "Successfully joined %s" % args[0]
            return response
        else:
            return "usage: join <channel>"

    def leaveChan(self, args):

        if len(args) > 0:
            if args[0] not in self.__channels:
                return "warning: not in %s" % args[0]
            response = ""
            if len(args) == 1:
                self.__channels[args[0]].leave()
            else:
                self.__channels[args[0]].leave(args[1]) 
                if not args[1:] == []:
                    response += "warning: \"%s\" garbage\n" % " ".join(args[1:])
            del self.__channels[args[0]]
            response += "Successfully leaved %s" % args[0]
            return response
        else:
            return "usage: leave <channel> [message]"

    def incomingChan(self, dirname):
        root = self.getIrcPath()
        if os.path.isdir(root + dirname):
            if os.path.exists(root + dirname + "/in") and os.path.isfile(root + dirname + "/out"):
                chan = Channel(dirname, self, connected=True)
                chan.join()
                self.__channels[dirname] = chan

    def watchChan(self):
        root = self.getIrcPath()
        while not self.__quit:
            chanlist = next(os.walk(self.getIrcPath()))[1]
            usrlist = list(filter(lambda k: not k[0] is "#" and not k in self.__treated, chanlist))
            #diff = list(set(usrlist) - set(self.__treated))
            for chan in usrlist:
                print("*** popped chan: " + chan)
                self.incomingChan(chan)
                self.__treated += [chan]
            time.sleep(0.1)

    def send_msg(self, connection, msg):
        # Prefix each message with a 4-byte length (network byte order)
        msg = struct.pack('>I', len(msg)) + msg
        connection.sendall(msg)

    def recv_msg(self, connection):
        # Read message length and unpack it into an integer
        raw_msglen = self.recvall(connection, 4)
        if not raw_msglen:
            return b''
        msglen = struct.unpack('>I', raw_msglen)[0]
        # Read the message data
        return self.recvall(connection, msglen)

    def recvall(self, connection,n):
        # Helper function to recv n bytes or return None if EOF is hit
        data = b''
        while len(data) < n:
            packet = connection.recv(n - len(data))
            if not packet:
                return b''
            data += packet
        return data

    def listenCommands(self):
        self.__sock.listen(1)
        while not self.__quit:
            connection, client = self.__sock.accept()
            try:
                command = self.recv_msg(connection).decode('UTF-8')
                print("*** command from control socket: %s" % command)
                self.parseCommand(command, connection)
            finally:
                connection.close()

    def parseCommand(self, command, connection):
        cmd = command.split(" ")
        orders = {
                "join"    : { "method": self.newChan,     "desc": "join a new channel" },
                "leave"   : { "method": self.leaveChan,   "desc": "leave an existing channel" },
                "quit"    : { "method": self.quit,        "desc": "quit sexiibot instance" },
                "say"     : { "method": self.messageChan, "desc": "say a message on a channel" },
                "action"  : { "method": self.actionChan,  "desc": "send an action message on a channel" },
                "nick"    : { "method": self.setNick,     "desc": "get or change nick" },
                }
        if cmd[0] == "help":
            message = "Welcome to sexiibot control help. Here is the list of available commands:\n"
            for item in orders:
                message += "  %s: %s\n" % (item, orders[item]["desc"])
            message += "  help: displays this help\n\n"
            message += "To have specific help for a command, just use it without any argument.\n"
            message += "Sexiibot loves you."
            self.send_msg(connection, bytes(message, 'UTF-8'))
        else:
            self.send_msg(connection, bytes(orders[cmd[0]]["method"](cmd[1:]), 'UTF-8'))

    def quit(self, args=[]):
        for chan in self.__channels:
            self.leaveChan([chan, "sexiibot stopping"])
        self.__quit = True
        self.__watch.join()
        self.__in.close()
        self.__out.close()
